@@ -1,11 +1,14 @@
 from difflib import unified_diff, Differ
 import json
+import shutil
 import sys
+from tempfile import mkdtemp
 import uuid
 import argparse
 import subprocess
 from contextlib import contextmanager
 from bashutils.colors import color_text
+import inject
 from mcloud.application import Application
 from mcloud.attach import AttachStdinProtocol
 from mcloud.config import YamlConfig
@@ -19,11 +22,11 @@ import pprintpp
 from prettytable import PrettyTable, ALL
 from texttable import Texttable
 from twisted.internet import defer, stdio
-from twisted.internet.defer import inlineCallbacks, CancelledError
+from twisted.internet.defer import inlineCallbacks, CancelledError, returnValue
 from twisted.internet.error import ConnectionRefusedError
 from mcloud import metadata
 import yaml
-
+from twisted.internet import reactor
 
 def format_epilog():
     """Program entry point.
@@ -52,10 +55,10 @@ arg_parser = argparse.ArgumentParser(
     formatter_class=argparse.RawDescriptionHelpFormatter,
     description=metadata.description,
     epilog=format_epilog(),
-    add_help=False
+    add_help=True
 )
 arg_parser.add_argument('-v', '--verbose', help='Show more logs', action='store_true', default=False)
-arg_parser.add_argument('-h', '--host', help='Host to use', default=None)
+arg_parser.add_argument('--host', help='Host to use', default=None)
 arg_parser.add_argument(
     '-V', '--version',
     action='version',
@@ -66,10 +69,73 @@ subparsers = arg_parser.add_subparsers()
 command_settings = {}
 
 
+class LocalCommand(object):
+    def __init__(self, config, command):
+        self.config = config
+        self.command = command
+
+    @inlineCallbacks
+    def call(self, to=None, **kwargs):
+
+        try:
+            host = config.get_command_host(name=to)
+        except KeyError:
+            raise Exception('Host alias not found.')
+
+        settings = inject.instance('settings')
+
+        uuid_ = uuid.uuid1()
+
+        for line in self.command['commands']:
+
+            line = line.replace('{host}', host)
+            line = line.replace('{uuid}', uuid_)
+
+            print(color_text(line, color='white', bcolor='blue'))
+
+            params = line.split(' ')
+            args = arg_parser.parse_args(params)
+
+            args.argv0 = sys.argv[0]
+
+            client = ApiRpcClient(host='127.0.0.1', settings=settings)
+
+            if isinstance(args.func, str):
+                yield getattr(client, args.func)(**vars(args))
+            else:
+                yield args.func(**vars(args))
+
+
+
+def load_commands(config):
+    """
+    :param config:
+    :type config: mcloud.config.YamlConfig
+    :return:
+    """
+
+    for name, command in config.get_commands().items():
+
+        cmd_instance = LocalCommand(config, command)
+
+        cmd = subparsers.add_parser('*%s' % name, help=command['help'])
+        cmd.add_argument('to', help='Host to use with command', default=None, choices=config.hosts, nargs='?')
+        cmd.set_defaults(func=cmd_instance.call)
+
+
+if os.path.exists('mcloud.yml'):
+    config = YamlConfig(file='mcloud.yml')
+    config.load(process=False)
+
+    load_commands(config)
+
+
+
+
 def cli(help_, arguments=None, by_ref=False, name=None):
     def cmd_decorator(func):
 
-        cmd = subparsers.add_parser(name or func.__name__, help=help_)
+        cmd = subparsers.add_parser(name or func.__name__.replace('_', '-'), help=help_)
 
         command_settings[func.__name__] = {
             'need_app': False
@@ -121,6 +187,8 @@ class ClientProcessInterruptHandler(object):
                 yield sleep(0.05)
 
 
+
+
 class ApiRpcClient(object):
     def __init__(self, host='127.0.0.1', port=7080, settings=None):
         self.host = host
@@ -230,7 +298,15 @@ class ApiRpcClient(object):
 
         x = PrettyTable(["Service name", "status", "ip", "cpu %", "memory", "volumes", "public urls"], hrules=ALL)
 
+        if app['status'] == 'error':
+            print ''
+            print 'Some errors occurred when receiving application information:'
+            for service in app['services']:
+                print '\n ' + service['name'] + ':'
+                print '  - ' + service['error']
+
         for service in app['services']:
+
             if service['created']:
                 service_status = 'ON' if service['running'] else 'OFF'
             else:
@@ -249,7 +325,7 @@ class ApiRpcClient(object):
                 if 'public_urls' in app and app['public_urls']:
                     for target in app['public_urls']:
                         url_ = target['url'] + '/'
-                        if target['port']:
+                        if 'port' in target and target['port']:
                             url_ += ' -> :' + target['port']
 
                         if not url_.startswith('http://'):
@@ -278,7 +354,7 @@ class ApiRpcClient(object):
 
     def print_app_list(self, data):
 
-        x = PrettyTable(["Application name", "status", "cpu %", "memory", "Web", "Path"], hrules=ALL)
+        x = PrettyTable(["Application name", "deployment", "status", "cpu %", "memory", "Web", "Path"], hrules=ALL)
         x.align = 'l'
         for app in data:
 
@@ -318,12 +394,19 @@ class ApiRpcClient(object):
                         if not url_.startswith('http://'):
                             url_ = 'http://' + url_
                         if 'service' in target and target['service']:
-                            web += '\n' + '%s -> [%s:%s]' % (url_, target['service'], target['port'] or '')
+                            web += '\n' + '%s -> [%s:%s]' % (url_, target['service'], target['port'] if 'port' in target else '')
                         else:
                             if web_service_:
                                 web += '\n' + '%s -> [%s]' % (url_, web_service_)
+            else:
+                app_status = '!error!'
 
-            x.add_row([app['name'], app_status, services_cpu_list, services_memory_list, web, app['config']['path']])
+
+            app_deployment = app['config']['deployment']
+            app_path = app['config']['path']
+
+            x.add_row([app['name'], app_deployment, app_status, services_cpu_list, services_memory_list, web,
+                       app_path])
 
         return '\n' + str(x) + '\n'
 
@@ -444,6 +527,13 @@ class ApiRpcClient(object):
     ############################################################
 
 
+    @cli('Execute mcloud shell', arguments=(
+        arg('shell', help='Continuously run list command'),
+    ))
+    @inlineCallbacks
+    def shell(self, **kwargs):
+        pass  # handled without argparser
+
     @cli('List registered applications', arguments=(
         arg('-f', '--follow', default=False, action='store_true', help='Continuously run list command'),
     ))
@@ -479,11 +569,12 @@ class ApiRpcClient(object):
     @cli('Creates a new application', arguments=(
         arg('ref', help='Application and service name', default=None, nargs='?'),
         arg('path', help='Path', nargs='?', default='.'),
+        arg('--deployment', help='Deployment', default=None),
         arg('--env', help='Application environment'),
         arg('--config', help='Config to use', default=None),
     ))
     @inlineCallbacks
-    def init(self, ref, path, config=None, env=None, **kwargs):
+    def init(self, ref, path, config=None, env=None, deployment=None, **kwargs):
 
         app, service = self.parse_app_ref(ref, kwargs, app_only=True)
 
@@ -495,12 +586,36 @@ class ApiRpcClient(object):
         config = YamlConfig(file=config_file, app_name=app)
         config.load(process=False)
 
-        if self.host != '127.0.0.1':
-            success = yield self._remote_exec('init', app, config=config.export(), env=env)
-            if success:
-                yield self.sync(path, '%s@%s' % (app, self.host), no_remove=False, force=True, full=True)
+        deployment_info = yield self._remote_exec('deployment_info', name=deployment)
+
+        if deployment_info:
+            if not deployment:
+                deployment = deployment_info['name']
+
+            print deployment_info
+
+            if not deployment_info['local']:
+                yield self._remote_exec('init', app, config=config.export(), env=env, deployment=deployment)
+            else:
+                yield self._remote_exec('init', app, path=os.path.realpath(path), config=config.export(), deployment=deployment)
+
         else:
-            yield self._remote_exec('init', app, path=os.path.realpath(path), config=config.export())
+            print('There is no deployments configured yet.\n\n'
+                  'You can create new local deployment using following command:\n'
+                  '\n  $ mcloud deployment-create local\n\n')
+
+            reactor.stop()
+
+    def represent_ordereddict(self, dumper, data):
+        value = []
+
+        for item_key, item_value in data.items():
+            node_key = dumper.represent_data(item_key)
+            node_value = dumper.represent_data(item_value)
+
+            value.append((node_key, node_value))
+
+        return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)
 
     @cli('Application configuration', arguments=(
         arg('ref', help='Application and service name', default=None, nargs='?'),
@@ -521,8 +636,9 @@ class ApiRpcClient(object):
             old_config = YamlConfig(source=unicode(app_config['source']), app_name=app, env=parser_env)
             old_config.load(process=False)
             from collections import OrderedDict
+
             yaml.add_representer(unicode, yaml.representer.SafeRepresenter.represent_unicode)
-            yaml.add_representer(OrderedDict, yaml.representer.SafeRepresenter.represent_dict)
+            yaml.add_representer(OrderedDict, self.represent_ordereddict)
             olds = yaml.dump(old_config.config, default_flow_style=False)
 
         if not update and not diff and not set_env:
@@ -543,6 +659,8 @@ class ApiRpcClient(object):
             new_config.load(process=False)
 
             if diff:
+                yaml.add_representer(unicode, yaml.representer.SafeRepresenter.represent_unicode)
+                yaml.add_representer(OrderedDict, self.represent_ordereddict)
                 news = yaml.dump(new_config.config, default_flow_style=False)
 
                 if olds == news:
@@ -576,15 +694,27 @@ class ApiRpcClient(object):
         data = yield self._remote_exec('remove', self.format_app_srv(app, service))
         print 'result: %s' % pprintpp.pformat(data)
 
+    @cli('Set app deployment', arguments=(
+        arg('ref', help='Application name', default=None, nargs='?'),
+        arg('deployment', help='Deployment name', default=None, nargs='?'),
+    ))
+    @inlineCallbacks
+    def set_deployment(self, ref, deployment, **kwargs):
+        app, service = self.parse_app_ref(ref, kwargs, app_only=True)
+        yield self._remote_exec('set_deployment', app, deployment)
+
+        yield self.list()
+
     ############################################################
 
     @cli('Start containers', arguments=(
         arg('ref', help='Application and service name', default=None, nargs='?'),
         arg('--init', help='Initialize applications if not exist yet', default=False, action='store_true'),
         arg('--env', help='Application environment'),
+        arg('--deployment', help='Application deployment'),
     ))
     @inlineCallbacks
-    def start(self, ref, init=False, env=None, **kwargs):
+    def start(self, ref, init=False, env=None, deployment=None, **kwargs):
 
         app, service = self.parse_app_ref(ref, kwargs)
 
@@ -592,7 +722,8 @@ class ApiRpcClient(object):
             app_instance = yield self.get_app(app)
 
             if not app_instance:
-                yield self.init(app, os.getcwd(), env=env)
+                yield self.init(app, os.getcwd(), env=env, deployment=deployment)
+                yield self.sync(os.getcwd(), '%s@%s' % (app, self.host), no_remove=False, force=True, full=True)
 
 
         data = yield self._remote_exec('start', self.format_app_srv(app, service))
@@ -614,6 +745,7 @@ class ApiRpcClient(object):
 
     @cli('Destroy containers', arguments=(
         arg('ref', help='Application and service name', default=None, nargs='?'),
+        arg('--scrub-data', default=False, action='store_true', help='Force volumes destroy'),
     ))
     @inlineCallbacks
     def destroy(self, ref, **kwargs):
@@ -637,6 +769,7 @@ class ApiRpcClient(object):
 
     @cli('Rebuild application', arguments=(
         arg('ref', help='Application and service name', default=None, nargs='?'),
+        arg('--scrub-data', default=False, action='store_true', help='Force volumes destroy'),
     ))
     @inlineCallbacks
     def rebuild(self, ref, **kwargs):
@@ -695,14 +828,110 @@ class ApiRpcClient(object):
 
     ############################################################
 
+
+    ############################################################
+    # Deployment life-cycle
+    ############################################################
+
+    @cli('List deployments')
+    @inlineCallbacks
+    def deployments(self, **kwargs):
+        data = yield self._remote_exec('deployments')
+
+        x = PrettyTable(["Deployment name", "Default", "host", "port", "local", "tls", "keys (ca|cert|key)"], hrules=ALL)
+        x.align = 'l'
+        for line in data:
+            certs = '|'.join(['x' if line[k] else 'o' for k in ('ca', 'cert', 'key')])
+            default = '*' if line['default'] else ''
+            x.add_row([line['name'], default, line['host'], line['port'], line['local'], line['tls'], certs])
+        print str(x)
+
+
+    @cli('Create deployment', arguments=(
+        arg('deployment', help='Deployment name'),
+        arg('ip_host', help='Deployment docker host', default=None, nargs='?'),
+        arg('--port', help='Deployment docker port', default=None),
+        arg('--tls', default=False, action='store_true', help='Use tls protocol'),
+        arg('--remote', default=True, dest='local', action='store_false', help='Deployment is remote'),
+    ))
+    @inlineCallbacks
+    def deployment_create(self, deployment, ip_host=None,  port=None, tls=None, local=True, **kwargs):
+        data = yield self._remote_exec('deployment_create', name=deployment, host=ip_host, port=port, tls=tls, local=local)
+
+        yield self.deployments()
+
+    @cli('update deployment', arguments=(
+        arg('deployment', help='Deployment name'),
+        arg('ip_host', help='Deployment docker host', default=None, nargs='?'),
+        arg('--port', help='Deployment docker port', default=None),
+        arg('--local', default=None, action='store_true', dest='local', help='Deployment is local'),
+        arg('--remote', default=None, action='store_false', dest='local', help='Deployment is remote'),
+        arg('--tls', action='store_true', dest='tls', help='Use tls protocol'),
+        arg('--no-tls', action='store_false', dest='tls', help='Don\'t use tls protocol'),
+    ))
+    @inlineCallbacks
+    def deployment_update(self, deployment, ip_host=None,  port=None, tls=None, local=None, **kwargs):
+        data = yield self._remote_exec('deployment_update', name=deployment, host=ip_host, port=port, tls=tls, local=local)
+
+        yield self.deployments()
+
+    @cli('make deployment default', arguments=(
+        arg('deployment', help='Deployment name'),
+    ))
+    @inlineCallbacks
+    def deployment_set_default(self, deployment, host, **kwargs):
+        yield self._remote_exec('deployment_set_default', name=deployment)
+
+        yield self.deployments()
+
+    @cli('upload key file to deployment', arguments=(
+        arg('deployment', help='Deployment name'),
+        arg('--ca', default=False, action='store_true', help='Upload ca'),
+        arg('--cert', default=False, action='store_true', help='Upload cert'),
+        arg('--key', default=False, action='store_true', help='Upload key'),
+        arg('--remove', default=False, action='store_true', help='Remove key'),
+        arg('infile', type=argparse.FileType('r'), default=sys.stdin, nargs='?'),
+    ))
+    @inlineCallbacks
+    def deployment_set(self, deployment, ca=False, cert=False, key=False, infile=None, remove=None, **kwargs):
+        if remove:
+            file_data = False
+        else:
+            file_data = infile.read()
+
+        data = {}
+
+        if ca:
+            data['ca'] = file_data
+
+        if cert:
+            data['cert'] = file_data
+
+        if key:
+            data['key'] = file_data
+
+        yield self._remote_exec('deployment_update', name=deployment, **data)
+
+        yield self.deployments()
+
+    @cli('remove deployment', arguments=(
+        arg('deployment', help='Deployment name'),
+    ))
+    @inlineCallbacks
+    def deployment_remove(self, deployment, **kwargs):
+        data = yield self._remote_exec('deployment_remove', name=deployment)
+        print 'result: %s' % pprintpp.pformat(data)
+
+        yield self.deployments()
+
     @cli('Publish an application', arguments=(
-        arg('ref', help='Application name', default=None, nargs='?'),
+        arg('ref', help='Application name'),
         arg('domain', help='Domain to publish'),
         arg('--port', help='Custom target port'),
         arg('--ssl', default=False, action='store_true', help='Ssl protocol'),
     ))
     @inlineCallbacks
-    def publish(self, domain, ref, ssl=False, port=None, **kwargs):
+    def publish(self, ref, domain, ssl=False, port=None, **kwargs):
         app_name, service = self.parse_app_ref(ref, kwargs, require_app=True)
 
         app = yield self.get_app(app_name)
@@ -710,16 +939,23 @@ class ApiRpcClient(object):
         if not app:
             print 'App not found. Can\'t publish'
         else:
-            yield self._remote_exec('publish', self.format_domain(domain, ssl), app_name, service, port)
+            data = yield self._remote_exec('publish', domain_name=self.format_domain(domain, ssl),
+                                    app_name=app_name, service_name=service, custom_port=port)
+
+            self.print_app_list(data)
 
     @cli('Unpublish an application', arguments=(
+        arg('ref', help='Application name'),
         arg('domain', help='Domain to unpublish'),
         arg('--ssl', default=False, action='store_true', help='Ssl protocol'),
     ))
     @inlineCallbacks
-    def unpublish(self, domain, ssl=False, **kwargs):
-        data = yield self._remote_exec('unpublish', self.format_domain(domain, ssl))
-        print 'result: %s' % pprintpp.pformat(data)
+    def unpublish(self, ref, domain, ssl=False, **kwargs):
+        app_name, service = self.parse_app_ref(ref, kwargs, require_app=True)
+
+        data = yield self._remote_exec('unpublish', app_name=app_name, domain_name=self.format_domain(domain, ssl))
+        self.print_app_list(data)
+
 
     ############################################################
     # Service utilities
@@ -796,7 +1032,7 @@ class ApiRpcClient(object):
     def push(self, volume, host, **kwargs):
         app, service = self.parse_app_ref(None, kwargs, app_only=True)
         config = yield self._remote_exec('config', app)
-        print config
+
 
 
     @cli('Syncronize application volumes', arguments=(
@@ -819,8 +1055,36 @@ class ApiRpcClient(object):
         elif src_type == 'local' and dst_type == 'remote':
             yield rsync_folder(self, dst_args, src_args, reverse=True, options=kwargs)
 
-        else:
-            print('%s to %s is not supported' % (src_type, dst_type))
+        elif src_type == 'remote' and dst_type == 'remote':
+            dir_name = mkdtemp()
+            dir_storage = get_storage(dir_name)
+
+            yield rsync_folder(self, src_args, dir_storage, options=kwargs)
+            yield rsync_folder(self, dst_args, dir_storage, reverse=True, options=kwargs)
+
+            shutil.rmtree(dir_name)
+
+    @cli('Backup application volumes', arguments=(
+        arg('source', help='source'),
+        arg('volume', help='Volume to backup', default=None),
+        arg('destination', help='Destination s3 bucket', default=None)
+    ))
+    @inlineCallbacks
+    def backup(self, source, volume, destination, **kwargs):
+        app_name, service = self.parse_app_ref(source, kwargs, require_app=True, require_service=True)
+        result = yield self._remote_exec('backup', app_name, service, volume, destination)
+        print result
+
+    @cli('Restore application volumes', arguments=(
+        arg('source', help='source'),
+        arg('volume', help='Volume to backup', default=None),
+        arg('destination', help='Destination s3 bucket', default=None)
+    ))
+    @inlineCallbacks
+    def restore(self, source, volume, destination, **kwargs):
+        app_name, service = self.parse_app_ref(source, kwargs, require_app=True, require_service=True)
+        result = yield self._remote_exec('backup', app_name, service, volume, destination, True)
+        print result
 
 
     ############################################################
