@@ -1,36 +1,79 @@
 import logging
+import traceback
+from mcloud.plugin import enumerate_plugins
+import re
 import inject
 from mcloud.remote import ApiRpcServer
-from mcloud.txdocker import IDockerClient, DockerConnectionFailed
+from mcloud.txdocker import IDockerClient, DockerConnectionFailed, DockerTwistedClient
 from twisted.internet import defer
 from twisted.internet.defer import inlineCallbacks, returnValue
 import txredisapi
+import os
+from zope.interface import Interface
+from zope.interface.verify import verifyObject
 
 logger = logging.getLogger('mcloud.application')
 
+
 class NotInspectedYet(Exception):
     pass
 
 
+class IServiceBuilder(Interface):
+    """
+    Allow plugins to participate in container build process.
+    """
+
+    def configure_container_on_start(service, config):
+        """
+        introspect service state and make changes on config dictionary that is passed to Docker
+        """
+
+    def configure_container_on_create(service, config):
+        """
+        introspect service state and make changes on config dictionary that is passed to Docker
+        """
+
+
+class IServiceLifecycleListener(Interface):
+
+    def on_service_start(service, ticket_id=None):
+        """
+        Called when service is started
+        """
+
+
 class Service(object):
 
     NotInspectedYet = NotInspectedYet
 
-    client = inject.attr(IDockerClient)
-
-    dns_server = inject.attr('dns-server')
+    settings = inject.attr('settings')
     dns_search_suffix = inject.attr('dns-search-suffix')
     redis = inject.attr(txredisapi.Connection)
 
     rpc_server = inject.attr(ApiRpcServer)
 
-    def __init__(self, **kwargs):
+    plugins = inject.attr('plugins')
+
+    error = False
+
+    def __init__(self, client=None, **kwargs):
+
+        self.client = client
+        """
+        @type client: DockerTwistedClient
+        """
+
         self.image_builder = None
         self.name = None
         self.app_name = None
+        self.entrypoint = None
+        self.workdir = None
         self.volumes = []
         self.volumes_from = None
         self.ports = None
+        self.web_port = None
+        self.ssl_port = None
         self.command = None
         self.env = None
         self.config = None
@@ -48,7 +91,10 @@ class Service(object):
 
 
     def task_log(self, ticket_id, message):
-        self.rpc_server.task_progress(message, ticket_id)
+        if not ticket_id:
+            print(message)
+        else:
+            self.rpc_server.task_progress(message, ticket_id)
 
     def build_docker_config(self):
         pass
@@ -57,15 +103,14 @@ class Service(object):
     @inlineCallbacks
     def inspect(self):
         self._inspected = True
-        data = yield self.client.inspect(self.name)
-        self._inspect_data = data
 
         try:
-            metrics = yield self.redis.hget('metrics', self.name)
-            if metrics:
-                self.memory_usage, self.cpu_usage = metrics.split(';')
-        except inject.InjectorException:
-            pass
+            data = yield self.client.inspect(self.name)
+            self._inspect_data = data
+
+        except DockerConnectionFailed as e:
+            self.error = 'Can not connect to docker: %s. %s' % (self.client.url, e)
+            self._inspect_data = None
 
         defer.returnValue(self._inspect_data)
 
@@ -77,6 +122,13 @@ class Service(object):
         except KeyError:
             return False
 
+    @property
+    def id(self):
+        if not self.is_created():
+            return None
+
+        return self._inspect_data['Id']
+
     @property
     def shortname(self):
         name_ = self.name
@@ -109,7 +161,7 @@ class Service(object):
         volumes_ = self._inspect_data['Volumes']
 
         internal_volumes = (
-            '/var/run/mcloud',
+            # '/var/run/mcloud',
             '/usr/bin/@me'
         )
         for iv in internal_volumes:
@@ -137,14 +189,19 @@ class Service(object):
         return self._inspect_data['State']['StartedAt']
 
     def is_web(self):
-        if not self.is_running():
-            return None
+        return not self.web_port is None
 
-        if not 'NetworkSettings' in self._inspect_data or not 'Ports' in self._inspect_data['NetworkSettings'] or \
-            not self._inspect_data['NetworkSettings']['Ports']:
-            return False
+    def get_web_port(self):
+        return self.web_port
 
-        return u'80/tcp' in self._inspect_data['NetworkSettings']['Ports']
+    def is_ssl(self):
+        return not self.ssl_port is None
+
+    def error_msg(self):
+        return self.error
+
+    def get_ssl_port(self):
+        return self.ssl_port
 
     def is_created(self):
         if not self.is_inspected():
@@ -152,10 +209,16 @@ class Service(object):
 
         return not self._inspect_data is None
 
+    def is_internal_volume(self, x):
+        return x == '/usr/bin/@me'
+
+    def is_read_only(self, x):
+        return self.is_internal_volume(x)
+
     @inlineCallbacks
     def run(self, ticket_id, command, size=None):
 
-        image_name = yield self.image_builder.build_image(ticket_id=ticket_id)
+        image_name = yield self.image_builder.build_image(ticket_id=ticket_id, service=self)
 
         config = yield self._generate_config(image_name, for_run=True)
 
@@ -169,18 +232,18 @@ class Service(object):
 
         yield self.client.create_container(config, name, ticket_id=ticket_id)
 
-        run_config = {
-            "Dns": [self.dns_server],
-            "DnsSearch": '%s.%s' % (self.app_name, self.dns_search_suffix)
-        }
-
-        run_config['VolumesFrom'] = self.name
+        run_config = {'VolumesFrom': self.name}
 
         if self.ports:
-            run_config['PortBindings'] = dict([(port, [{}]) for port in self.ports])
+            config['PortBindings'] = self.prepare_ports()
 
         if self.volumes and len(self.volumes):
-            run_config['Binds'] = ['%s:%s' % (x['local'], x['remote']) for x in self.volumes]
+            run_config['Binds'] = ['%s:%s' % (x['local'], x['remote'] + (':ro' if self.is_read_only(x['remote']) else '')) for x in self.volumes]
+
+
+
+        for plugin in enumerate_plugins(IServiceBuilder):
+            yield plugin.configure_container_on_start(self, run_config)
 
         yield self.client.start_container(name, ticket_id=ticket_id, config=run_config)
 
@@ -190,8 +253,26 @@ class Service(object):
         else:
             yield self.client.attach(name, ticket_id, skip_terminal=True)
 
+    def prepare_ports(self):
+        all_ports = {}
+        for port in self.ports:
+            if isinstance(port, basestring) and ':' in port:
+                local, host = port.split(':')
+                if '_' in host:
+                    ip, rport = host.split('_')
+                else:
+                    rport = host
+                    ip = None
+
+                all_ports[local] = [{"HostPort": rport, "HostIp": ip}]
+            else:
+                all_ports[port] = [{}]
+
+        return all_ports
+
     @inlineCallbacks
-    def start(self, ticket_id):
+    def start(self, ticket_id=None):
+
         id_ = yield self.client.find_container_by_name(self.name)
 
         self.task_log(ticket_id, '[%s][%s] Starting service' % (ticket_id, self.name))
@@ -203,28 +284,66 @@ class Service(object):
             yield self.create(ticket_id)
             id_ = yield self.client.find_container_by_name(self.name)
 
+        current_config = yield self.inspect()
+        image_id = current_config['Image']
+        image_info = yield self.client.inspect_image(image_id)
+
         self.task_log(ticket_id, '[%s][%s] Starting service...' % (ticket_id, self.name))
 
         config = {
-            "Dns": [self.dns_server],
-            "DnsSearch": '%s.%s' % (self.app_name, self.dns_search_suffix)
+            # "Dns": [self.dns_server],
+            # "DnsSearch": '%s.%s' % (self.app_name, self.dns_search_suffix)
         }
 
+        for plugin in enumerate_plugins(IServiceBuilder):
+            yield plugin.configure_container_on_start(self, config)
+
+
         if self.volumes_from:
             config['VolumesFrom'] = self.volumes_from
 
         if self.ports:
-            config['PortBindings'] = dict([(port, [{}]) for port in self.ports])
+            config['PortBindings'] = self.prepare_ports()
 
+        mounted_volumes = []
+        config['Binds'] = []
         if self.volumes and len(self.volumes):
-            config['Binds'] = ['%s:%s' % (x['local'], x['remote']) for x in self.volumes]
+            for x in self.volumes:
+                mounted_volumes.append(x['remote'])
+                config['Binds'].append('%s:%s' % (x['local'], x['remote'] + (':ro' if self.is_read_only(x['remote']) else '')))
 
-        self.task_log(ticket_id, 'Startng container with config: %s' % config)
+
+        if image_info['ContainerConfig']['Volumes']:
+            for vpath, vinfo in image_info['ContainerConfig']['Volumes'].items():
+
+                if not vpath in mounted_volumes:
+                    dir_ = os.path.expanduser('%s/volumes/%s/%s' % (self.settings.home_dir, self.name, re.sub('[^a-z0-9]+', '_', vpath)))
+
+                    if self.settings.btrfs:
+                        dir_ += '_btrfs'
+
+                    if not os.path.exists(dir_):
+                        if self.settings.btrfs:
+                            os.system('btrfs subvolume create %s' % dir_)
+                        else:
+                            os.makedirs(dir_)
+
+                    mounted_volumes.append(vpath)
+                    config['Binds'].append('%s:%s' % (dir_, vpath))
 
         #config['Binds'] = ["/home/alex/dev/mcloud/examples/static_site1/public:/var/www"]
 
+        self.task_log(ticket_id, 'Startng container. config: %s' % config)
+
         yield self.client.start_container(id_, ticket_id=ticket_id, config=config)
 
+        # lifecycle events
+
+        self.task_log(ticket_id, 'Emit startup event')
+        for plugin in enumerate_plugins(IServiceLifecycleListener):
+            self.task_log(ticket_id, 'Call start listener %s' % plugin)
+            yield plugin.on_service_start(self, ticket_id=ticket_id)
+
         # inspect and return result
         ret = yield self.inspect()
 
@@ -232,7 +351,20 @@ class Service(object):
 
 
     @inlineCallbacks
-    def stop(self, ticket_id):
+    def restart(self, ticket_id=None):
+        yield self.stop(ticket_id)
+        yield self.start(ticket_id)
+
+
+    @inlineCallbacks
+    def rebuild(self, ticket_id=None):
+        yield self.stop(ticket_id)
+        yield self.destroy(ticket_id)
+        yield self.start(ticket_id)
+
+
+    @inlineCallbacks
+    def stop(self, ticket_id=None):
 
         id = yield self.client.find_container_by_name(self.name)
 
@@ -241,12 +373,47 @@ class Service(object):
         ret = yield self.inspect()
         defer.returnValue(ret)
 
+
+    @inlineCallbacks
+    def pause(self, ticket_id=None):
+
+        id = yield self.client.find_container_by_name(self.name)
+
+        yield self.client.pause_container(id, ticket_id=ticket_id)
+
+        ret = yield self.inspect()
+        defer.returnValue(ret)
+
+
+    @inlineCallbacks
+    def unpause(self, ticket_id=None):
+
+        id = yield self.client.find_container_by_name(self.name)
+
+        yield self.client.unpause_container(id, ticket_id=ticket_id)
+
+        ret = yield self.inspect()
+        defer.returnValue(ret)
+
+
+    @inlineCallbacks
+    def last_logs(self, on_log):
+        id = yield self.client.find_container_by_name(self.name)
+
+        yield self.client.logs(self, id, on_log, tail=100, follow=False)
+
     @inlineCallbacks
     def _generate_config(self, image_name, for_run=False):
         config = {
             "Image": image_name,
+            # "RestartPolicy": "always"
         }
 
+        image_info = None
+        # TODO: improve tests
+        if hasattr(self.client, 'inspect_image'):
+            image_info = yield self.client.inspect_image(image_name)
+
         vlist = yield self.redis.hgetall('vars')
 
         if self.env:
@@ -255,9 +422,23 @@ class Service(object):
         if len(vlist) > 0:
             config['Env'] = ['%s=%s' % x for x in vlist.items()]
 
-        if self.ports:
-            config['ExposedPorts'] = dict([(port, {}) for port in self.ports])
 
+        if self.ports:
+            all_ports = {}
+            for port in self.ports:
+                if isinstance(port, basestring) and ':' in port:
+                    all_ports[port.split(':')[0]] = {}
+                else:
+                    all_ports[port] = {}
+
+            config['ExposedPorts'] = all_ports
+
+
+        if self.entrypoint:
+            config['Entrypoint'] = self.entrypoint
+
+        if self.workdir:
+            config['WorkingDir'] = self.workdir
 
         if not for_run:
             config['Hostname'] = self.name
@@ -270,13 +451,19 @@ class Service(object):
                     (x['remote'], {}) for x in self.volumes
                 ])
 
+            if image_info and image_info['ContainerConfig']['Volumes']:
+                for vpath, vinfo in image_info['ContainerConfig']['Volumes'].items():
+                    config['Volumes'][vpath] = {}
+
+        for plugin in enumerate_plugins(IServiceBuilder):
+            yield plugin.configure_container_on_create(self, config)
+
         defer.returnValue(config)
 
 
     @inlineCallbacks
-    def create(self, ticket_id):
-
-        image_name = yield self.image_builder.build_image(ticket_id=ticket_id)
+    def create(self, ticket_id=None):
+        image_name = yield self.image_builder.build_image(ticket_id=ticket_id, service=self)
 
         config = yield self._generate_config(image_name)
         yield self.client.create_container(config, self.name, ticket_id=ticket_id)
@@ -285,8 +472,9 @@ class Service(object):
         defer.returnValue(ret)
 
     @inlineCallbacks
-    def destroy(self, ticket_id):
+    def destroy(self, ticket_id=None):
         id_ = yield self.client.find_container_by_name(self.name)
+
         yield self.client.remove_container(id_, ticket_id=ticket_id)
 
         ret = yield self.inspect()
