@@ -1,11 +1,15 @@
 import inspect
 import random
 import string
+import uuid
 from mcloud.container import PrebuiltImageBuilder
+
 from mcloud.service import Service
+
 from mcloud.sync import VolumeNotFound
 import os
 
+
 import re
 from autobahn.twisted.util import sleep
 import inject
@@ -14,14 +18,50 @@ from twisted.internet.defer import inlineCallbacks
 from twisted.internet.error import ConnectionDone
 import txredisapi
 from mcloud.txdocker import IDockerClient, NotFound
+
+
 from mcloud.application import ApplicationController, AppDoesNotExist
 from mcloud.deployment import DeploymentController
 from mcloud.events import EventBus
 from mcloud.remote import ApiRpcServer
 
+from twisted.internet import protocol
+
+
+class TicketScopeProcess(protocol.ProcessProtocol):
+    def __init__(self, ticket_id, client):
+        self.ticket_id = ticket_id
+        self.client = client
+        self.d = None
+
+    def call_sync(self, *args, **kwargs):
+        self.log('Executing process with args: %s' % str(args))
+        self.d = defer.Deferred()
+        reactor.spawnProcess(self, *args, **kwargs)
+
+        return self.d
+
+    def log(self, data):
+        self.client.task_log(self.ticket_id, data)
+
+    def connectionMade(self):
+        self.transport.closeStdin()
+
+    def outReceived(self, data):
+        self.log(data)
+
+    def errReceived(self, data):
+        self.log(data)
+
+    def processExited(self, reason):
+        self.log("processExited, status %d" % (reason.value.exitCode,))
+
+    def processEnded(self, reason):
+        self.log("processEnded, status %d" % (reason.value.exitCode,))
+        self.d.callback(True)
+
 
 class TaskService(object):
-
     app_controller = inject.attr(ApplicationController)
     """
     @type app_controller: ApplicationController
@@ -32,18 +72,16 @@ class TaskService(object):
     event_bus = inject.attr(EventBus)
     """ @type: EventBus """
 
-    dns_server = inject.attr('dns-server')
-    dns_search_suffix = inject.attr('dns-search-suffix')
+    settings = inject.attr('settings')
 
     def task_log(self, ticket_id, message):
-
         self.rpc_server.task_progress(message, ticket_id)
 
     def task_help(self, ticket_id):
         pass
 
     @inlineCallbacks
-    def task_init(self, ticket_id, name, path=None, config=None, env=None):
+    def task_init(self, ticket_id, name, path=None, config=None, env=None, deployment=None):
         """
         Initialize new application
 
@@ -53,6 +91,9 @@ class TaskService(object):
         :return:
         """
 
+        if not deployment:
+            Exception('Deployment name is required!')
+
         app = None
         try:
             app = yield self.app_controller.get(name)
@@ -66,12 +107,9 @@ class TaskService(object):
             raise ValueError('config must be provided to create an application')
 
         if not path:
-            home_dir = os.path.expanduser('~/.mcloud')
-            path = os.path.join(home_dir, name)
-            if not os.path.exists(path):
-                os.makedirs(path, 0700)
+            path = '/root/.mcloud/%s' % name
 
-        yield self.app_controller.create(name, {'path': path, 'source': config, 'env': env})
+        yield self.app_controller.create(name, {'path': path, 'source': config, 'env': env, 'deployment': deployment})
 
         defer.returnValue(True)
 
@@ -165,9 +203,23 @@ class TaskService(object):
         ret = 'Done.'
         defer.returnValue(ret)
 
+    @inlineCallbacks
+    def task_set_deployment(self, ticket_id, app, deployment):
+        """
+        Remove application
+
+        :param ticket_id:
+        :param name:
+        :return:
+        """
+        yield self.app_controller.update(app, {'deployment': deployment})
+
+        ret = 'Done.'
+        defer.returnValue(ret)
+
 
     @inlineCallbacks
-    def task_logs(self, ticket_id, name):
+    def task_logs(self, ticket_id, ref):
         """
         Read logs.
 
@@ -178,34 +230,24 @@ class TaskService(object):
         :return:
         """
 
+        service, app = ref.split('.')
+
+        # todo: fix to remote client
+
         def on_log(log):
+
             if len(log) == 8 and log[7] != 0x0a:
                 return
 
             self.task_log(ticket_id, log)
 
         try:
-            client = inject.instance(IDockerClient)
-            yield client.logs(name, on_log, tail=100)
-        except NotFound:
-            self.task_log(ticket_id, 'Container not found by name.')
+            app = yield self.app_controller.get(app)
 
+            config = yield app.load()
 
-    @inlineCallbacks
-    def task_attach(self, ticket_id, container_id, size=None):
-        """
-        Attach to container.
-
-        TaskIO is attached to container.
-
-        :param ticket_id:
-        :param container_id:
-        :param size:
-        :return:
-        """
-        try:
-            client = inject.instance(IDockerClient)
-            yield client.attach(container_id, ticket_id)
+            service = config.get_service(ref)
+            yield service.client.logs(service.name, on_log, tail=100)
         except NotFound:
             self.task_log(ticket_id, 'Container not found by name.')
 
@@ -231,9 +273,7 @@ class TaskService(object):
 
             config = yield app.load()
 
-            services = config.get_services()
-
-            service = services['%s.%s' % (service_name, app_name)]
+            service = config.get_service('%s.%s' % (service_name, app_name))
 
             yield service.run(ticket_id, command, size=size)
 
@@ -260,7 +300,7 @@ class TaskService(object):
             'path': app.config['path'],
             'env': app.get_env(),
             'source': app.config['source'] if 'source' in app.config else {},
-            'hosts': config.get_hosts() if hasattr(config, 'get_hosts') else {},
+            'hosts': config.hosts,
             'volumes': config.get_volumes() if hasattr(config, 'get_volumes') else {},
         })
 
@@ -303,18 +343,6 @@ class TaskService(object):
         return d
 
 
-    @inlineCallbacks
-    def task_dns(self, ticket_id):
-        """
-        Show dns allocation table.
-
-        :param ticket_id:
-        :return:
-        """
-        data = yield self.redis.hgetall('domain')
-
-        defer.returnValue(data)
-
     @inlineCallbacks
     def task_restart(self, ticket_id, name):
         """
@@ -347,22 +375,25 @@ class TaskService(object):
 
     def follow_logs(self, service, ticket_id):
 
-
         def on_log(log):
+
+            if log.startswith('@mcloud ready in '):
+                parts = log.split(' ')
+                self.event_bus.fire_event('api.%s.%s' % (service.name, 'ready'), my_args=parts[2:])
+                return
+
             if len(log) == 8 and log[7] != 0x0a:
                 return
 
             self.task_log(ticket_id, log)
 
         def done(result):
-           pass
+            pass
 
         def on_err(failure):
-           pass
+            print failure
 
-        client = inject.instance(IDockerClient)
-
-        d = client.logs(service.name, on_log)
+        d = service.client.logs(service.name, on_log)
         d.addCallback(done)
         d.addErrback(on_err)
 
@@ -373,7 +404,12 @@ class TaskService(object):
 
     @inlineCallbacks
     def task_sync_stop(self, ticket_id, app_name, sync_ticket_id):
-        s = Service()
+
+        app = yield self.app_controller.get(app_name)
+
+        client = yield app.get_client()
+
+        s = Service(client=client)
         s.app_name = app_name
         s.name = '%s_%s_%s' % (app_name, '_rsync_', sync_ticket_id)
 
@@ -394,9 +430,9 @@ class TaskService(object):
         app = yield self.app_controller.get(app_name)
 
         config = yield app.load()
+        client = yield app.get_client()
 
-
-        s = Service()
+        s = Service(client=client)
         s.app_name = app_name
         s.name = '%s_%s_%s' % (app_name, '_rsync_', ticket_id)
         s.image_builder = PrebuiltImageBuilder(image='modera/rsync')
@@ -425,28 +461,108 @@ class TaskService(object):
 
         else:
             s.volumes = [{
-                'local': app.config['path'],
-                'remote': '/volume'
-            }]
+                             'local': app.config['path'],
+                             'remote': '/volume'
+                         }]
             volume_name = '/volume'
 
         s.env = {
             'USERNAME': ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(32)),
-            'PASSWORD': ''.join(random.choice(string.ascii_lowercase + string.punctuation + string.digits) for _ in range(32)),
+            'PASSWORD': ''.join(
+                random.choice(string.ascii_lowercase + string.punctuation + string.digits) for _ in range(32)),
             'ALLOW': '*'
         }
 
         yield s.start(ticket_id)
 
+        deployment = yield app.get_deployment()
+
+        if deployment.local:
+            sync_host = 'me'
+        else:
+            sync_host = deployment.host
+
         defer.returnValue({
             'env': s.env,
             'container': s.name,
+            'host': sync_host,
             'port': s.public_ports()['873'][0]['HostPort'],
             'volume': volume_name,
             'ticket_id': ticket_id
         })
 
 
+    @inlineCallbacks
+    def task_backup(self, ticket_id, app_name, service_name, volume, destination, restore=False):
+
+        app = yield self.app_controller.get(app_name)
+
+        config = yield app.load()
+
+        service = None
+
+        if service_name:
+
+            if not volume:
+                raise VolumeNotFound('In case of service name is provided, volume name is mandatory!')
+
+            services = config.get_services()
+
+            service_full_name = '%s.%s' % (service_name, app_name)
+            try:
+                service = services[service_full_name]
+
+                all_volumes = service.list_volumes()
+
+                volume_path = all_volumes[volume]
+
+            except KeyError:
+                raise VolumeNotFound('Service with name %s was not found!' % service_name)
+
+        else:
+            volume_path = app.config['path']
+
+        if not restore:
+
+            if self.settings.btrfs:
+                uuid_ = uuid.uuid1()
+
+                snapshot_path = '%s/snapshots_%s' % (self.settings.home_dir, uuid_)
+
+                self.task_log(ticket_id, snapshot_path)
+
+                yield TicketScopeProcess(ticket_id, self).call_sync(
+                    '/sbin/btrfs', ['btrfs', 'subvolume', 'snapshot', '-r', volume_path, snapshot_path]
+                )
+                self.task_log(ticket_id, '-----------------')
+
+                volume_path = snapshot_path
+
+            else:
+                if service:
+                    service.pause()
+
+            yield TicketScopeProcess(ticket_id, self).call_sync(
+                'aws', ['aws', 's3', 'sync', volume_path, destination]
+            )
+
+            if not self.settings.btrfs:
+                service.unpause()
+            else:
+                yield TicketScopeProcess(ticket_id, self).call_sync(
+                    '/sbin/btrfs', ['btrfs', 'subvolume', 'delete', volume_path]
+                )
+
+        else:
+            yield TicketScopeProcess(ticket_id, self).call_sync(
+                'aws', ['aws', 's3', 'sync', destination, volume_path]
+            )
+
+        defer.returnValue({
+            'status': 'ok',
+            'path': volume_path
+        })
+
 
     @inlineCallbacks
     def task_start(self, ticket_id, name):
@@ -523,24 +639,28 @@ class TaskService(object):
                         yield service.inspect()
 
                         if not service.is_running():
-                            self.task_log(ticket_id, 'FATAL: Service is not running after timeout. Stopping application execution.')
+                            self.task_log(ticket_id,
+                                          'FATAL: Service is not running after timeout. Stopping application execution.')
                             log_process.cancel()
                             defer.returnValue(False)
                         else:
                             self.task_log(ticket_id, 'Container still up. Continue execution.')
                     else:
                         sleep_time = 0.5
+                        print event
                         if 'my_args' in event and len(event['my_args']) == 2:
                             if event['my_args'][0] == 'in':
                                 match = re.match('^([0-9]+)s$', event['my_args'][1])
                                 if match:
                                     sleep_time = float(match.group(1))
 
-                        self.task_log(ticket_id, 'Container is waiting %ss to make sure container is started.' % sleep_time)
+                        self.task_log(ticket_id,
+                                      'Container is waiting %ss to make sure container is started.' % sleep_time)
                         yield sleep(sleep_time)
 
                         if not service.is_running():
-                            self.task_log(ticket_id, 'FATAL: Service is not running after ready report. Stopping application execution.')
+                            self.task_log(ticket_id,
+                                          'FATAL: Service is not running after ready report. Stopping application execution.')
                             log_process.cancel()
                             defer.returnValue(False)
                         else:
@@ -765,24 +885,59 @@ class TaskService(object):
         ret = yield defer.gatherResults(deployment_list, consumeErrors=True)
         defer.returnValue(ret)
 
-    #
-    # @inlineCallbacks
-    # def task_deployment_create(self, ticket_id, public_domain):
-    # deployment = yield self.deployment_controller.create(public_domain)
-    #     defer.returnValue(not deployment is None)
-    #
-    # @inlineCallbacks
-    # def task_deployment_new_app_zip(self, ticket_id, deployment_name, name, path):
-    #     app = yield self.deployment_controller.new_app(deployment_name, name, {'path': path})
-    #     defer.returnValue(not app is None)
-    #
-    # @inlineCallbacks
-    # def task_deployment_new_app_source(self, ticket_id, deployment_name, name, source):
-    #     app = yield self.deployment_controller.new_app(deployment_name, name, {'source': source})
-    #     defer.returnValue(not app is None)
 
     @inlineCallbacks
-    def task_publish(self, ticket_id, deployment_name, app_name, service_name, custom_port=None):
+    def task_deployment_info(self, ticket_id, name=None):
+        if name:
+            deployment = yield self.deployment_controller.get(name=name)
+        else:
+            deployment = yield self.deployment_controller.get_default()
+
+        if not deployment:
+            defer.returnValue(None)
+
+        data = yield deployment.load_data()
+
+        defer.returnValue(data)
+
+
+    @inlineCallbacks
+    def task_app_deployment_info(self, ticket_id, name=None):
+        app = yield self.app_controller.get(name=name)
+        deployment = yield app.get_deployment()
+
+        if not deployment:
+            defer.returnValue(None)
+
+        data = yield deployment.load_data()
+
+        defer.returnValue(data)
+
+    @inlineCallbacks
+    def task_deployment_create(self, ticket_id, **kwargs):
+        deployment = yield self.deployment_controller.create(**kwargs)
+        deployments = yield self.deployment_controller.list()
+        if len(deployments) == 1:
+            yield self.deployment_controller.set_default(deployment.name)
+            
+        defer.returnValue(not deployment is None)
+
+    @inlineCallbacks
+    def task_deployment_set_default(self, ticket_id, name):
+        yield self.deployment_controller.set_default(name)
+
+    @inlineCallbacks
+    def task_deployment_update(self, ticket_id, **kwargs):
+        deployment = yield self.deployment_controller.update(**kwargs)
+        defer.returnValue(not deployment is None)
+
+    @inlineCallbacks
+    def task_deployment_remove(self, ticket_id, name):
+        deployment = yield self.deployment_controller.remove(name)
+        defer.returnValue(deployment is None)
+
+    @inlineCallbacks
+    def task_publish(self, ticket_id, domain_name, app_name, service_name, custom_port=None):
         """
         Publish application URL.
 
@@ -791,13 +946,16 @@ class TaskService(object):
         :param app_name:
         :return:
         """
-        yield self.deployment_controller.publish_app(deployment_name, app_name, service_name, custom_port)
+        app = yield self.app_controller.get(app_name)
+        deployment = yield app.get_deployment()
+
+        yield self.deployment_controller.publish_app(deployment, domain_name, app_name, service_name, custom_port, ticket_id=ticket_id)
 
         ret = yield self.app_controller.list()
         defer.returnValue(ret)
 
     @inlineCallbacks
-    def task_unpublish(self, ticket_id, deployment_name):
+    def task_unpublish(self, ticket_id, domain_name, app_name):
         """
         Unpublish URL
 
@@ -805,8 +963,10 @@ class TaskService(object):
         :param deployment_name:
         :return:
         """
+        app = yield self.app_controller.get(app_name)
+        deployment = yield app.get_deployment()
 
-        yield self.deployment_controller.unpublish_app(deployment_name)
+        yield self.deployment_controller.unpublish_app(deployment, domain_name, ticket_id=ticket_id)
 
         ret = yield self.app_controller.list()
         defer.returnValue(ret)
